/******** TODO ************
 *  
 *  
 *  Make useful profile presets
 *  
 *  Working On:
 *    
 *  
 *  DONE/History:
 *  011: added dpad
 *  010: added option to allow output to Titan2 during edit or start (profile selection) additionally
 *  saving/loading to/from eeprom
 *  Save profile changes, load profiles
 *  008: option for button name/number format
 *  Press anything to close Start Logo
 *  007: Autoincrease of selected values in mapping editing when holding down the display-plus or -minus button
 *  Enable editing of selected profile
 *  005: Backup for incomplete but working profile editing
 *  005: changed editing mode 
 *  003: Mode M_START - Profile selection / entering+leaving edit mode / entering+leaving profile selection mode
 *  Added small graphics for better button position recognisation
 *  Added automatic switching between upper and lower buttons (controller and display)
 */


/* Input buttons to keyboard / joystick mapping

   You must select Keyboard+Mouse+Joystick from the "Tools > USB Type" menu
   
   // Button->Teensy-Pin connections

   [a]           [f]  // rear buttons 
    23            0

   [b]           [g] // rear buttons
    22            1

 {c][d][e]   {h][i][j] // stick buttons
 17|20|21     2  3  4
 
       _________
      |         |     // display & display buttons
   [k]|         |[l]
   16 |_________| 9
     [m]  [n]  [o]
     15    8    10
  */

#include <Bounce.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <EEPROM.h>
#include <stdio.h>

#define VERSION "1.0.0"

// defines of the snap on controller name -> button array index number for "easier" understanding
#define BTN_R4  0  // right upper rear button
#define BTN_R5  1  // right lower rear button
#define BTN_RS1 2  // around right stick
#define BTN_RS2 3  // around right stick
#define BTN_RS3 4  // around right stick
#define BTN_DE  5  // display-edit / select (below display middle)
#define BTN_DR  6  // right of display
#define BTN_DP  7  // display-plus (below screen, right)
#define BTN_DM  8  // display-minus (below screen, left)
#define BTN_DL  9  // left of display
#define BTN_LS1 10 // around left stick
#define BTN_LS2 11 // around left stick
#define BTN_LS3 12 // around left stick
#define BTN_L4  13 // left upper rear button
#define BTN_L5  14 // left lower rear button

uint16_t JoyHat=0;
byte DUp=0;
byte DDown=0;
byte DLeft=0;
byte DRight=0;

#define BUTTONS_UPPER 1 // buttons on the back and around the analog sticks
#define BUTTONS_LOWER 2 // buttons around the display
#define SHOWTIME_INTRO 8000 // msec for showing the intro screen
#define SHOWTIME_UPPER 5000 // msec for showing the upper button mapping screen
#define SHOWTIME_LOWER 5000 // msec for showing the lower button mapping screen
#define SHOW true  // for showing the content of a screen
#define HIDE false  // for redrawing the screen content in black, hiding the content again

#define M_START   0 // system start - displaying last used profile mapping for some time
byte    M_START_Step = 0; // current step in start screen mode
#define M_MANUAL  1 // user input during M_START , result depends on button pressed: Display-Left /-Right = switch profiles , Display-Bottom-Center = disable screen -> sets mode M_RUNNING
#define M_EDIT    2 // entering editing mode
byte    M_EDIT_Step = 0; // current step in editing
byte    M_EDIT_Btn = 0; // button id to edit / show in detail
byte    M_EDIT_Btn_Last = 0; // last button id that was edited / shown in detail
byte    M_EDIT_Modify = 1; // current step when modifying a button
#define M_RUN     3 // profile is selected / accepted - show "Profile #1 activated" for some time (and turn screen of ?)
#define M_RUNNING 4 // silently running the profile ( turned screen off ? )
#define M_OPTION 5 // handling the option menu
byte    M_OPTION_Step = 0; // current step in options
byte    M_OPTION_Modify = 1; // current step when modifying an option

#define DISPLAY_TOGGLE_BTN 8 // which button to use for turning display on/off
#define DISPLAY_ON_TIME 1000  // how many msec the button has to be held down for reactivating the screen
#define DISPLAY_OFF_TIME 1000  // how many msec the button has to be held down for disabling the screen
uint32_t DisplayToggleTimer = 0; // keeping track of the time the button is hold down

#define EDIT_TOGGLE_BTN 5 // which button to use for entering / leaving profile editing mode
#define EDIT_ON_TIME 1000  // how many msec the button has to be held down for entering edit mode
#define EDIT_OFF_TIME 1000  // how many msec the button has to be held down for leaving edit mode
uint32_t EditToggleTimer = 0; // keeping track of the time the button is hold down

#define OPTION_TOGGLE_BTN BTN_DP // which button to use for entering / leaving options
#define OPTION_ON_TIME 1000  // how many msec the button has to be held down for entering this mode
#define OPTION_OFF_TIME 1000  // how many msec the button has to be held down for leaving this mode
uint32_t OptionToggleTimer = 0; // keeping track of the time the button is hold down

#define DOUBLETAPTIME 300 // the time limit for a doubletap
uint32_t DoubleTapTimer = 0; // keeping track of the time the button was last pressed

#define AUTOINC_WAIT 500 // the time to wait before starting automatic increasing of selected values (buttons/letters in button mapping editing)
#define AUTOINC_SPEED 100 // the time to wait before speeding up the automatic increase/decrease the next step
uint32_t AutoIncWaitTimer = 0; // // keeping track of the time the button was last pressed
uint32_t AutoIncTimer = 0; // keeping track of the time the next valu increase should happen
bool AutoIncUp = 0; // 1 = up, 0 = down



const int imax = 15; // maximum of hardware buttons
const int PROFILEMAX = 4; // maximum of profiles

//                             0   1   2   3   4   5   6    7   8   9  10  11  12  13  14   // <- Position in array
const byte ibuttons[imax] = {  6,  7,  2,  3,  4,  8,  9,  10, 15, 16, 17, 20, 21, 22, 23}; // button connected pins
//const byte lbuttons[imax] = {  1,  1,  2,  2,  2,  3,  3,   3,  3,  3,  2,  2,  2,  1,  1}; // location of the buttons: 0 = not available, 1 = rear, 2 = around Sticks, 3 = around Display

//  Snap On Device button names:  0     1     2      3      4      5     6     7     8     9     10     11     12     13    14   // <- Position in array
const char nbuttons[imax][4] = { "R4", "R5", "RS1", "RS2", "RS3", "DE", "DR", "DO", "DX", "DL", "LS1", "LS2", "LS3", "L5", "L4"};

const char T2BTNS[2][22][12]  = {
                             // PS4
                            { "PS4","PS", "Share", "Options",                       // 0, 1, 2, 3
                              "R1", "R2", "R3", "L1", "L2", "L3",                // 4, 5, 6, 7, 8, 9
                              "DPad Up", "DPad Down", "DPad Left", "DPad Right", // 10, 11, 12, 13
                              "Triangle", "Circle", "Cross", "Square",           // 14, 15, 16, 17
                              "Touch Click", "Touch P1", "noname", "Touch P2"    // 18, 19, 20, 21
                            },

                            // XBOX One Elite                             
                             { "XBox1", "Xbox", "View", "Menu",                       // 0, 1, 2, 3
                              "RB", "RT", "RS", "LB", "LT", "LS",                // 4, 5, 6, 7, 8, 9
                              "DPad Up", "DPad Down", "DPad Left", "DPad Right", // 10, 11, 12, 13
                              "Y", "B", "A", "X",                                // 14, 15, 16, 17
                              "P1", "P2", "P3", "P4"                             // 18, 19, 20, 21
                            }
};
byte T2BtnNaming = 0; // 0 = PS4 DS$, 1 = Xbox One Elite

// 0=show none,1=show number for T2 output, 2=show number for SnapT input, 3=show both, 4=show debug SnapT input and unmapped raw T2 number output) | eeprom pos: 2
#define SNAPT_BTNNR_MAX 5 // count of possible formats (see below)
const char SnapTBTNNRS[SNAPT_BTNNR_MAX][9] = { "None", "T2 Out", "SnapT In", "Both", "Debug"};

// Teensy Button Output Numbers IDs > T2 Buttons IDs,
//  meaning: when teensy send Button 0 as pressed, Windows/T2 receives it as 17
//                             0, 1,  2,  3, 4, 5, 6, 7, 8,  9, 10, 11, 12, 13, 14, 15, 16, 17, v18, v19, v20, v21 // v = virtual : for mapping joystick.hat
const byte te2t2buttons[] = { 0, 17, 16, 15, 14, 7, 4, 8, 5, 2,  3,  9,  6,  1, 18, 19, 20, 21,  10, 11, 12, 13}; // TODO: special handling of 10..13 required for DPad
// the other way around
const byte t22tebuttons[] = { 0, 17, 16, 15, 14, 7, 4, 8, 5, 2,  3,  9,  6,  1, 18, 19, 20, 21,  10, 11, 12, 13}; // : DPAD  10..13 (pos 18..21) (joystick.hat)

struct Profile {
  //                       0   1   2   3   4   5   6    7   8   9  10  11  12  13  14   // <- Position in array
  bool tbuttons[imax] = {  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0}; // 0 = keyboard press, 1 = joystick button
  char kbuttons[imax] = { 'f','g','h','i','j','n','l', 'o','m','k','c','d','e','b','a'}; // assign characters / keys to snap-on buttons
  byte jbuttons[imax] = {  6,  7,  8,  9,  10, 14, 12,  15, 13, 11, 3,  4,  5,  2,  1}; // assign joystick buttons to snap-on buttons
};
Profile Profiles[PROFILEMAX];

// eeprom settings
byte ProfileActive = 0; // active/lastused profile number | eeprom pos: 0
byte EE_ButtonNames = 0; // button naming scheme (0=ps4,1=xbox1) | eeprom pos: 1
byte EE_ButtonNumber = 0; // show button numbers  (0=show none,1=show number for T2 output, 2=show number for SnapT input, 3=show both, 4=show debug SnapT input and unmapped raw T2 number output) | eeprom pos: 2
byte EE_OutputOnModes = 0; // when to send button/keypresses to the Titan Two (0=Run,1=+Profile,2=+Edit,3=+Profile+Edit) | eeprom pos: 3
// Profile data in between 10 and 200


bool DisplayOn = true; // when the display should be on (displaying some data)
bool RefreshScreen = false; // when the screen should be redrawn
byte ShowData = BUTTONS_UPPER; // byte ShowData = BUTTONS_LOWER;
uint32_t tNow = millis(); // actual system time in msec
uint32_t tNextScreen = tNow + SHOWTIME_INTRO; // for automatic switching between upper/lower button mapping overview

bool joyupd = false; // only when joystick button state changed send an update

// handling different program states, like startup, editing, running,...
byte Mode = M_START;

#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);

#if (SSD1306_LCDHEIGHT != 64)
#error("Height incorrect, please fix Adafruit_SSD1306.h!");
#endif

//------------------------------------------------------------------------------
// File generated by LCD Assistant
// http://en.radzio.dxp.pl/bitmap_converter/
// Consoltuner Titan Two Logo, thanks to J2Kbr for allowing to use it here
//------------------------------------------------------------------------------
const unsigned char PROGMEM OLedTitanTwo [] = {
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x18, 0xE3, 0x62, 0x30, 0x00, 0x8C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0E,
0x18, 0xE3, 0x06, 0x30, 0x01, 0x8C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18,
0x0C, 0xE6, 0x6F, 0xB7, 0x03, 0xED, 0xC1, 0xC0, 0x36, 0x07, 0x18, 0x8C, 0x70, 0xD8, 0x0E, 0x3E,
0x0D, 0xF6, 0x6F, 0xBF, 0x83, 0xEF, 0xE3, 0xE0, 0x3F, 0x0F, 0x99, 0xCC, 0xF8, 0xF8, 0x1F, 0x3E,
0x0D, 0xB6, 0x66, 0x39, 0x81, 0x8E, 0x66, 0x30, 0x3B, 0x9D, 0xCD, 0xD9, 0x8C, 0xE0, 0x3B, 0x98,
0x0D, 0xB6, 0x66, 0x31, 0x81, 0x8C, 0x67, 0xF0, 0x31, 0x98, 0xCD, 0xD9, 0xFC, 0xC0, 0x31, 0x98,
0x0F, 0xBE, 0x66, 0x31, 0x81, 0x8C, 0x67, 0xF0, 0x31, 0x98, 0xCF, 0x79, 0xFC, 0xC0, 0x31, 0x98,
0x07, 0x1C, 0x66, 0x31, 0x81, 0x8C, 0x66, 0x00, 0x3B, 0x9D, 0xC7, 0x71, 0x80, 0xC0, 0x3B, 0x98,
0x07, 0x1C, 0x67, 0xB1, 0x81, 0xEC, 0x63, 0xF0, 0x3F, 0x0F, 0x87, 0x70, 0xFC, 0xC0, 0x1F, 0x18,
0x07, 0x1C, 0x63, 0xB1, 0x80, 0xEC, 0x61, 0xE0, 0x36, 0x07, 0x06, 0x30, 0x78, 0xC0, 0x0E, 0x18,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x7F, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x03, 0xFF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x0F, 0xFF, 0xFC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x1F, 0xFF, 0xFE, 0x00, 0x0F, 0xFF, 0xFC, 0xEF, 0xFF, 0xFC, 0x7F, 0x00, 0x00, 0x00, 0x00,
0x00, 0x3F, 0xFF, 0xFF, 0x00, 0x0C, 0x00, 0x06, 0x18, 0x00, 0x02, 0xFF, 0x01, 0xFE, 0x0F, 0x80,
0x00, 0x1F, 0xFF, 0xFE, 0x00, 0x1F, 0xFF, 0xC4, 0x00, 0x00, 0x03, 0x00, 0x83, 0xFF, 0x1F, 0x80,
0x00, 0xCF, 0xFF, 0xFC, 0xC0, 0x3F, 0xFF, 0xF9, 0x80, 0x00, 0x02, 0x00, 0x82, 0x0D, 0x1F, 0x80,
0x01, 0xE7, 0xFF, 0xF9, 0xE0, 0x3F, 0xFF, 0xFF, 0xFF, 0x90, 0x04, 0x00, 0x40, 0x00, 0xA1, 0x80,
0x01, 0xF3, 0xFF, 0xF3, 0xE0, 0x00, 0xFF, 0x0F, 0xF0, 0xFF, 0x5B, 0x00, 0x44, 0x00, 0xE1, 0x00,
0x03, 0xF9, 0xFF, 0xE7, 0xF0, 0x00, 0xFE, 0x0F, 0xE0, 0xFE, 0x17, 0xFF, 0xC4, 0x00, 0x41, 0x00,
0x03, 0xFC, 0x3F, 0x0F, 0xF0, 0x00, 0xFE, 0x0F, 0xE0, 0xFE, 0x3F, 0xEF, 0xA3, 0xB4, 0x01, 0x00,
0x03, 0xFC, 0x3F, 0x0F, 0xF0, 0x01, 0xFC, 0x0F, 0xC0, 0xFC, 0x7F, 0xFF, 0xEF, 0xF7, 0x9A, 0x00,
0x07, 0xFC, 0x3F, 0x0F, 0xF8, 0x01, 0xFC, 0x17, 0x41, 0x74, 0xBF, 0xFF, 0xFF, 0xCF, 0xFA, 0x00,
0x07, 0xFC, 0x3F, 0x0F, 0xF8, 0x01, 0xFC, 0x10, 0x41, 0x44, 0x7B, 0x7F, 0xFF, 0xCF, 0xFE, 0x00,
0x07, 0xFC, 0x1F, 0x0F, 0xF8, 0x03, 0xF8, 0x20, 0x82, 0x0C, 0x04, 0xF8, 0x47, 0xC7, 0xFC, 0x00,
0x07, 0xF8, 0x1E, 0x07, 0xF8, 0x03, 0xF8, 0x1F, 0x83, 0xFB, 0xF8, 0x07, 0xFF, 0x87, 0xFC, 0x00,
0x07, 0xF8, 0x1E, 0x07, 0xF8, 0x01, 0xF0, 0x1F, 0x01, 0xF3, 0xF0, 0x03, 0xFF, 0x03, 0xF8, 0x00,
0x07, 0xFC, 0x1E, 0x0F, 0xF8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x07, 0xFC, 0x1E, 0x0F, 0xF8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x07, 0xFC, 0x0C, 0x0F, 0xF8, 0x00, 0x00, 0x00, 0x00, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xF8, 0x00,
0x03, 0xFE, 0x0C, 0x1F, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xF0, 0x00,
0x03, 0xFF, 0x0C, 0x3F, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xF0, 0x00,
0x01, 0xFF, 0x80, 0x7F, 0xE0, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x06, 0x31, 0x8C, 0x01, 0xE0, 0x00,
0x01, 0xFF, 0xF3, 0xFF, 0xE0, 0x00, 0x00, 0x00, 0x00, 0xF6, 0x3E, 0x31, 0x18, 0x01, 0xE0, 0x00,
0x01, 0xFF, 0xFF, 0xFF, 0xE0, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x7E, 0x63, 0x19, 0xF3, 0xE0, 0x00,
0x00, 0xFF, 0xFF, 0xFF, 0xC0, 0x00, 0x00, 0x00, 0x01, 0xFE, 0x7C, 0x63, 0x31, 0xE3, 0xC0, 0x00,
0x00, 0x7F, 0xFF, 0xFF, 0x80, 0x00, 0x00, 0x00, 0x01, 0xFC, 0x7C, 0x00, 0x30, 0x07, 0xC0, 0x00,
0x00, 0x3F, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x01, 0xFC, 0xFC, 0x00, 0x30, 0x0F, 0xC0, 0x00,
0x00, 0x1F, 0xFF, 0xFE, 0x00, 0x00, 0x00, 0x00, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x00,
0x00, 0x07, 0xFF, 0xF8, 0x00, 0x00, 0x00, 0x00, 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x00,
0x00, 0x03, 0xFF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x07, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x80, 0x00,
0x00, 0x00, 0x7F, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x11, 0x31, 0x31, 0x21, 0x87, 0x16, 0x18, 0x71, 0x1C, 0xE8, 0x96, 0x1C, 0x50, 0x61, 0xC5, 0x98,
0x13, 0x33, 0x33, 0x22, 0x48, 0x99, 0x24, 0x89, 0x22, 0x48, 0x99, 0x22, 0x60, 0x92, 0x26, 0x64,
0x15, 0x55, 0x55, 0x44, 0x10, 0x91, 0x31, 0x09, 0x42, 0x48, 0x91, 0x42, 0x41, 0x04, 0x24, 0x44,
0x15, 0x55, 0x55, 0x44, 0x10, 0x91, 0x0D, 0x09, 0x7E, 0x88, 0x91, 0x7E, 0x41, 0x04, 0x24, 0x44,
0x0D, 0x8D, 0x8D, 0x84, 0x10, 0x91, 0x45, 0x09, 0x40, 0x91, 0x11, 0x40, 0x81, 0x04, 0x28, 0x44,
0x09, 0x89, 0x89, 0x84, 0x91, 0x22, 0x45, 0x12, 0x44, 0x93, 0x22, 0x44, 0x81, 0x24, 0x48, 0x88,
0x09, 0x09, 0x09, 0x23, 0x0E, 0x22, 0x38, 0xE2, 0x38, 0xCD, 0x22, 0x38, 0x88, 0xC3, 0x88, 0x88,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

// Create Bounce objects for each button.  The Bounce object
// automatically deals with contact chatter or "bounce", and
// it makes detecting changes very simple.
Bounce bbuttons[] = {
  Bounce( 6,10 ), //  0
  Bounce( 7,10 ), //  1
  Bounce( 2,10 ), //  2
  Bounce( 3,10 ), //  3
  Bounce( 4,10 ), //  4
  Bounce( 8,10 ), //  5
  Bounce( 9,10 ), //  6
  Bounce( 10,10 ),//  7
  Bounce( 15,10 ),//  8
  Bounce( 16,10 ),//  9
  Bounce( 17,10 ),// 10
  Bounce( 20,10 ),// 11
  Bounce( 21,10 ),// 12
  Bounce( 22,10 ),// 13
  Bounce( 23,10 ),// 14
};

void setup() {
  Serial.begin(9600);
  
  // initializing the joystick data, only send data when we have to (should fix the Titan Two 'E4' error)
  Joystick.useManualSend(true);
  Joystick.X(512); // 512 is center
  Joystick.Y(512);
  Joystick.Z(512);
  Joystick.Zrotate(512);
  Joystick.sliderLeft(512);
  Joystick.sliderRight(512);
  Joystick.hat(-1);  // registered as button 10,11,12,13 on Titan Two depending of the hats directions (Up/Right/Down/Left)
  Joystick.send_now(); // send once so the system will have the chance to detect it as a gamepad/joystick
  
  // Configure the pins for input mode with pullup resistors.
  // The pushbuttons connect from each pin to ground.  When
  // the button is pressed, the pin reads LOW because the button
  // shorts it to ground.  When released, the pin reads HIGH
  // because the pullup resistor connects to +5 volts inside
  // the chip.  LOW for "on", and HIGH for "off" may seem
  // backwards, but using the on-chip pullup resistors is very
  // convenient.  The scheme is called "active low", and it's
  // very commonly used in electronics... so much that the chip
  // has built-in pullup resistors!
  for (int i=0; i <= imax-1; i++){
    pinMode(ibuttons[i], INPUT_PULLUP);
  }

  // Display
  DisplayOn = true;
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3C (for the 128x64 I2C)
  display.clearDisplay();
  display.drawBitmap(0, 0,  OLedTitanTwo, 128, 64, 1);
  display.display();
  LoadConfig(); // load settings, lastused profile and all profile data
}

void Edit_Screen_Draw(byte screen, byte draw, byte btn) {
  display.setTextSize(1);
  if ( draw == SHOW ) display.setTextColor(WHITE,BLACK);
  else display.setTextColor(BLACK,BLACK);
  switch(screen) {
    case 1: // entering editing mode info screen
            ShowProfileNr(true);
            display.setCursor(4,18);
            display.println("Editing Mode entered");
            display.setCursor(4,40);
            display.println("Press button to edit");
            break;
    case 2: // showing deailed info of the tapped button
            ShowProfileNr(true);
            display.setCursor(0,16);
            display.print("Button: ");
            if (EE_ButtonNumber == 2 || EE_ButtonNumber == 3 ) {
              display.print(btn);
              display.print(".");
            }
            display.print(nbuttons[btn]);

            display.setCursor(0,28);
            display.print("Function: ");
            if ( Profiles[ProfileActive].tbuttons[btn] == 0 ) {
              display.print("Keyboard");
              display.setCursor(0,40);
              display.print("Key: ");
              display.print(Profiles[ProfileActive].kbuttons[btn]);
            }
            else {
              display.print("Joypad");
              display.setCursor(0,40);
              display.print("Button: ");
              if (EE_ButtonNumber == 1 || EE_ButtonNumber == 3 ) {
                display.print(Profiles[ProfileActive].jbuttons[btn]);
                display.print(".");
              }
              else if ( EE_ButtonNumber == 4) { // debug : show full button number info : button ID as T2 : output translation for T2 to really act as this button ID
                display.print(Profiles[ProfileActive].jbuttons[btn]); // 1
                display.print(":");
                display.print(te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]); // 17
                display.print(".");
              }
              display.print(T2BTNS[T2BtnNaming][te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]]);
              /*
              Serial.print(Profiles[ProfileActive].jbuttons[btn]); // 1
              Serial.print(":");
              Serial.print(te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]); // 17
              Serial.print(".");
              Serial.print(T2BTNS[T2BtnNaming][te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]]); // PS , sollte aber Square sein !
              */
            }
            display.setCursor(0,56);
            display.print("! Doubletap to edit !");
            break;
    case 3: // showing modification screen of the mapping of the selected button
            display.clearDisplay();
            ShowProfileNr(true);
            display.setCursor(0,16);
            display.print("Button: ");
            if (EE_ButtonNumber == 2 || EE_ButtonNumber == 3 ) {
              display.print(btn);
              display.print(".");
            }
            display.print(nbuttons[btn]);

            display.setCursor(0,28);
            display.print("Function:");
            if ( Profiles[ProfileActive].tbuttons[btn] == 0 ) {
              if ( draw == SHOW && M_EDIT_Modify == 1) display.setTextColor(BLACK, WHITE); // 'inverted' text
              display.print(" ");
              display.print("Keyboard");
              display.print(" ");
              display.setCursor(0,40);
              if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
              display.print("Key:");
              if ( draw == SHOW && M_EDIT_Modify == 2) display.setTextColor(BLACK, WHITE); // 'inverted' text
              else if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
              display.print(" ");
              display.print(Profiles[ProfileActive].kbuttons[btn]);
              display.print(" ");
            }
            else {
              if ( draw == SHOW && M_EDIT_Modify == 1) display.setTextColor(BLACK, WHITE); // 'inverted' text
              else if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
              display.print(" ");
              display.print("Joypad");
              display.print(" ");
              display.setCursor(0,40);
              if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
              display.print("Button:");
              if ( draw == SHOW && M_EDIT_Modify == 2) display.setTextColor(BLACK, WHITE); // 'inverted' text
              else if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
              display.print(" ");
              if ( EE_ButtonNumber == 1 || EE_ButtonNumber == 3 ) {
                display.print(Profiles[ProfileActive].jbuttons[btn]);
                display.print(".");
              } else if ( EE_ButtonNumber == 4) { // debug : show full button number info : button ID as T2 : output translation for T2 to really act as this button ID
                display.print(Profiles[ProfileActive].jbuttons[btn]);
                display.print(":");
                display.print(te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]);
                display.print(".");
              }
              display.print(T2BTNS[T2BtnNaming][te2t2buttons[Profiles[ProfileActive].jbuttons[btn]]]);
              display.print(" ");
            }
            display.setCursor(40,56);
            if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
            if ( draw == SHOW && M_EDIT_Modify == 3) display.setTextColor(BLACK, WHITE); // 'inverted' text
            display.print("> Done <");
            break;    
  } // eo : switch(screen)
}

void Option_Screen_Draw(byte screen, byte draw) {
  display.setTextSize(1);
  if ( draw == SHOW ) display.setTextColor(WHITE,BLACK);
  else display.setTextColor(BLACK,BLACK);
  display.setCursor(0,0);
  display.print("Options");
  display.setCursor(80,0);
  display.print("v");
  display.print(VERSION);
  switch(screen) {
    case 1: // showing the option editing screen
            display.setCursor(0,16);
            if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
            display.print("Button-Names:");
            if ( draw == SHOW && M_OPTION_Modify == 1) display.setTextColor(BLACK, WHITE); // 'inverted' text
            display.print(" ");
            display.print(T2BTNS[EE_ButtonNames][0]);
            display.print(" ");
            
            display.setCursor(0,28);
            if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
            display.print("Button-Nr:");
            if ( draw == SHOW && M_OPTION_Modify == 2) display.setTextColor(BLACK, WHITE); // 'inverted' text
            display.print(" ");
            display.print(SnapTBTNNRS[EE_ButtonNumber]);
            display.print(" ");

            display.setCursor(0,40);
            if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
            display.print("Output On:");
            if ( draw == SHOW && M_OPTION_Modify == 3) display.setTextColor(BLACK, WHITE); // 'inverted' text
            display.print(" ");
            display.print("Run ");
            if ( EE_OutputOnModes == 1 ) {
              display.setCursor(60,48);
              display.print(" +Profile ");
            }
            if ( EE_OutputOnModes == 2 ) display.print("+Edit ");
            if ( EE_OutputOnModes == 3 ) {
              display.setCursor(35,48);
              display.print(" +Profile+Edit ");
            }
            display.setCursor(40,56);
            if ( draw == SHOW ) display.setTextColor(WHITE,BLACK); // normal text
            if ( draw == SHOW && M_OPTION_Modify == 4) display.setTextColor(BLACK, WHITE); // 'inverted' text
            display.print("> Done <");
  }
}

void loop() {
  static byte ActiveScreen = 0;
  tNow = millis();
  
  // startup mode - displaying lastused profile settings
  if ( Mode == M_START ) {
    switch (M_START_Step) {
      case 0: // initial startup screen
        if ( tNow >= tNextScreen ) { // switch to next
          M_START_Step = 1;
        }
        break;
      case 1: // upper / lower button layout
        if ( tNow >= tNextScreen ) { // switching between upper and lower button mapping screen every n msec
          RefreshScreen = true;
          tNextScreen = tNow + SHOWTIME_UPPER;
          ActiveScreen++;
          if ( ActiveScreen > 2 ) ActiveScreen = 1;
        }
        if ( RefreshScreen == true ) {
          ShowData = ActiveScreen;
          display.clearDisplay();
          display.setTextSize(1);
          display.setTextColor(WHITE);
          ShowProfileNr(false);
          ShowProfileData(ProfileActive);
          RefreshScreen = false;
        }
    }
  } // eo: M_START
  
  else if ( Mode == M_EDIT ) {
    switch (M_EDIT_Step) {
      case 1: // Waiting for button to edit the mapping
        display.clearDisplay();
        Edit_Screen_Draw(1,SHOW,0); // show "press button to edit" screen
        display.display();
        M_EDIT_Step = 2;
        break;
      case 3: // Show mapping selection screen
        Edit_Screen_Draw(1,HIDE,0); // removing text from step 1
        Edit_Screen_Draw(2,HIDE,M_EDIT_Btn_Last); // removing text from previous button
        Edit_Screen_Draw(2,SHOW,M_EDIT_Btn); // show button map detail screen
        display.display();
        M_EDIT_Step = 2; // back to edit mode / button selection
        break;
      case 4: // Show mapping editing screen
        Edit_Screen_Draw(2,HIDE,M_EDIT_Btn); // removing text from previous button
        Edit_Screen_Draw(3,SHOW,M_EDIT_Btn); // display editing text for button
        display.display();
        break;
    } 
  }// eo: M_EDIT
    
  else if ( Mode == M_OPTION ) {
    switch (M_OPTION_Step) {
      case 1: // Option Button-Naming PS4/XBOX1
      case 2: // Option Show Button Nr
      case 3: // Option Output during modes
      case 4: // Option leaving
        display.clearDisplay();
        Option_Screen_Draw(1,SHOW);
        display.display();
        break;
    }
  } // eo: M_OPTION
  
  // Update all the buttons.  There should not be any long
  // delays in loop(), so this runs repetitively at a rate
  // faster than the buttons could be pressed and released.
  joyupd = false;
  JoyHat=0; // helper variable for translating button press into Joystick.hat value to act as DPad
  for (int i=0; i < imax; i++){
    bbuttons[i].update();

    // Check each button for "falling" edge.
    // Type a message on the Keyboard when each button presses
    // Update the Joystick buttons only upon changes.
    // falling = high (not pressed - voltage from pullup resistor)
    //           to low (pressed - button connects pin to ground)
    if (bbuttons[i].fallingEdge()) {
      if ( Mode == M_START ) { // startup profile selection handling
        if ( EE_OutputOnModes==1 || EE_OutputOnModes==3 ) { // send output during start too ?
          // keyboard output handling
          if ( Profiles[ProfileActive].tbuttons[i] == 0 ) Keyboard.press(Profiles[ProfileActive].kbuttons[i]);
          else { // joypad handling
            JoyPadHandling(i,1);
          } // eo: joypad output handling
        }
        if ( M_START_Step == 0 ) {
          RefreshScreen = true;
          tNextScreen = tNow + SHOWTIME_UPPER;
          ActiveScreen = 1;
          M_START_Step = 1;
        } else {
          switch(i) {
            case BTN_DL : // switch to previous profile : left display button (Btn ID 9)
                      if ( ProfileActive > 0 ) ProfileActive--;
                      else ProfileActive = PROFILEMAX -1;
                      ActiveScreen = 0;
                      tNextScreen = 0;
                      break;
            case BTN_DR : // switch to next profile : right display button (Btn ID 6)
                      ProfileActive++;
                      if ( ProfileActive >= PROFILEMAX ) ProfileActive = 0;
                      ActiveScreen = 0;
                      tNextScreen = 0;
                      break;
            case EDIT_TOGGLE_BTN :  // entering edit mode ? && switch pad/screen map overview
                      RefreshScreen = true;
                      tNextScreen = tNow;
                      EditToggleTimer = tNow; // store the actual time
                      break;
            case DISPLAY_TOGGLE_BTN : // accept profile (and turn off screen ?)  && switch pad/screen map overview
                      RefreshScreen = true;
                      tNextScreen = tNow;
                      DisplayToggleTimer = tNow; // store the actual time
                      break;
            case OPTION_TOGGLE_BTN : // entering option mode  && switch pad/screen map overview
                      RefreshScreen = true;
                      tNextScreen = tNow;
                      OptionToggleTimer = tNow; // store the actual time
                      break;
          } // eo: switch i
        } // eo: else M_START_Step == 0
      }  // eo: falling edge - Mode == M_START

      else if ( Mode == M_EDIT ) {
        switch(i) {
                    case EDIT_TOGGLE_BTN :  // leave the editing mode
                    EditToggleTimer = tNow; // store the actual time
                    break;
        }
        
        if ( EE_OutputOnModes==2 || EE_OutputOnModes==3 ) { // send output during editing too ?
          // keyboard output handling
          if ( Profiles[ProfileActive].tbuttons[i] == 0 ) Keyboard.press(Profiles[ProfileActive].kbuttons[i]);
          else { // joypad handling
            JoyPadHandling(i,1);
          } // eo: joypad output handling
        }

        switch(M_EDIT_Step) {
          case 2: // button to edit was pressed
                  // check for doubletap of the same button to enter editing of it
                  if ( tNow - DoubleTapTimer <= DOUBLETAPTIME && M_EDIT_Btn_Last == M_EDIT_Btn ) {
                    M_EDIT_Modify = 1;
                    M_EDIT_Step = 4;
                  } else {
                    M_EDIT_Step = 3;
                  }
                  
                  DoubleTapTimer = tNow;
                  M_EDIT_Btn_Last = M_EDIT_Btn;
                  M_EDIT_Btn = i;
                  break;
          case 4: // modifying a button mapping
                  if ( i == BTN_DR ) { // display-right : next possible setting to modify
                      M_EDIT_Modify++;
                      if (M_EDIT_Modify > 3 ) M_EDIT_Modify = 1;
                  }
                  if ( i == BTN_DE ) { // display-edit (below display) : next possible setting to modify
                      if (M_EDIT_Modify < 3 ) M_EDIT_Modify++;
                      else if (M_EDIT_Modify == 3 ) {
                        Edit_Screen_Draw(3,HIDE,M_EDIT_Btn); // display editing text for button
                        M_EDIT_Step = 3; // "> Done <" is selected
                      }
                  }
                  if ( i == BTN_DL ) { // display-left : previous possible setting to modify
                      M_EDIT_Modify--;
                      if (M_EDIT_Modify < 1 ) M_EDIT_Modify = 3;
                  }
                  if ( i == BTN_DP || i == BTN_DM ) { // edit button, centered below the display || left of this button
                    if ( M_EDIT_Modify == 1 ) { // "Function" is selected
                      // erase old settings shown on display
                      Edit_Screen_Draw(3,HIDE,M_EDIT_Btn); // display editing text for button
                      // update setting
                      if (Profiles[ProfileActive].tbuttons[M_EDIT_Btn] == 0) Profiles[ProfileActive].tbuttons[M_EDIT_Btn]= 1;
                      else Profiles[ProfileActive].tbuttons[M_EDIT_Btn]= 0;
                    }
                    else if ( M_EDIT_Modify == 2 ) { // Button / key is selected
                      if (Profiles[ProfileActive].tbuttons[M_EDIT_Btn] == 0) { // keyboard
                        if ( i == BTN_DP ) { // next character 0..9, a..Z
                          AutoIncWaitTimer = tNow;
                          AutoIncUp = 1;
                          KButtonMappingMod(1);
                        } else if ( i == BTN_DM ) { // previous character  0..9, a..Z
                          AutoIncWaitTimer = tNow;
                          AutoIncUp = 0;
                          KButtonMappingMod(0);
                        }
                        
                      } else { // joypad
                        if ( i == BTN_DP ) { // next joypad button 0 .. 
                          AutoIncWaitTimer = tNow;
                          AutoIncUp = 1;
                          JButtonMappingMod(1);
                        } else if ( i == BTN_DM ) { // previous joypad button 0 ..
                          AutoIncWaitTimer = tNow;
                          AutoIncUp = 0;
                          JButtonMappingMod(0);
                        }
                      }
                                            
                    }
                    else if ( M_EDIT_Modify == 3 ) {
                      Edit_Screen_Draw(3,HIDE,M_EDIT_Btn); // display editing text for button
                      M_EDIT_Step = 3; // "> Done <" is selected
                    }
                  } // eo: if i == BTN_DP || BTN_DM
        } // eo : switch(M_EDIT_Step)
      }

      else if ( Mode == M_RUNNING ) { // input handling when profile is active and screen if off
        switch(i) {
          case DISPLAY_TOGGLE_BTN :   DisplayToggleTimer = tNow; // store the actual time
                                      break;
        }
      
        if ( Profiles[ProfileActive].tbuttons[i] == 0 ) Keyboard.press(Profiles[ProfileActive].kbuttons[i]);
        else { // joypad handling
          Joystick.button(Profiles[ProfileActive].jbuttons[i], 1);
          joyupd = true;
        } // eo: joypad handling
      } // eo: falling edge - Mode == M_RUNNING

      else if ( Mode == M_OPTION ) { // input handling for option menu
        if ( i == BTN_DE ) {
          if ( M_OPTION_Modify < 4 ) M_OPTION_Modify++;
          else {
            Mode = M_START;
            ActiveScreen = 0;
            tNextScreen = 0;
            SaveConfig();
          }
        }
        else if ( i == BTN_DR ) {
          M_OPTION_Modify++;
          if ( M_OPTION_Modify > 4 ) M_OPTION_Modify = 1;
        } 
        else if ( i == BTN_DL ) {
          M_OPTION_Modify--;
          if ( M_OPTION_Modify < 0 ) M_OPTION_Modify = 4;
        } 
        switch (M_OPTION_Modify) { // the different options
          case 4: if ( i == BTN_DP || i == BTN_DM ) {
                    Mode = M_START;
                    ActiveScreen = 0;
                    tNextScreen = 0;
                    SaveConfig();
                  }
          case 1 :  if ( i == BTN_DP || i == BTN_DM ) {
                      if ( EE_ButtonNames == 0 ) EE_ButtonNames++;
                      else if ( EE_ButtonNames == 1 ) EE_ButtonNames--;
                    }
                    break;
          case 2 : if ( i == BTN_DP ) {
                      if ( EE_ButtonNumber < SNAPT_BTNNR_MAX-1 ) EE_ButtonNumber++;
                      else EE_ButtonNumber = 0;
                    } 
                    else if ( i == BTN_DM ) {
                      if ( EE_ButtonNumber > 0 ) EE_ButtonNumber--;
                      else EE_ButtonNumber = SNAPT_BTNNR_MAX-1;
                    }
                    break;
          case 3:  if ( i == BTN_DP ) {
                      if ( EE_OutputOnModes < 3 ) EE_OutputOnModes++;
                      else if ( EE_OutputOnModes == 3 ) EE_OutputOnModes=0;
                    }
                    else if ( i == BTN_DM ) {
                      if ( EE_OutputOnModes > 0 ) EE_OutputOnModes--;
                      else if ( EE_OutputOnModes == 0 ) EE_OutputOnModes=3;
                    }
                    break;
        }
      } // eo: falling edge - Mode == M_OPTION
      
    } // eo: fallingEdge
  
      // Check each button for "rising" edge
      // Type a message on the Keyboard when each button releases.
      // For many types of projects, you only care when the button
      // is pressed and the release isn't needed.
      // rising = low (pressed - button connects pin to ground)
      //          to high (not pressed - voltage from pullup resistor)
      if (bbuttons[i].risingEdge()) {
        AutoIncWaitTimer = 0;
        switch(i) {
          case EDIT_TOGGLE_BTN :   EditToggleTimer = 0;
                                   break;
          case DISPLAY_TOGGLE_BTN: DisplayToggleTimer = 0;
                                   break;
          case OPTION_TOGGLE_BTN : OptionToggleTimer = 0;
                                   break;
        }
        
        if ( Profiles[ProfileActive].tbuttons[i] == 0 ) Keyboard.release(Profiles[ProfileActive].kbuttons[i]);
        else {
            JoyPadHandling(i,0);
            joyupd = true;
          }
      } // eo: risingEdge

  } // eo: button for loop
      
  if ( joyupd == true ) {
    //Serial.println(jhat);
    Joystick.send_now();
  }

  // autoinc button map editing values
  if ( AutoIncWaitTimer == 0 ) AutoIncTimer = 0;
  if ( (AutoIncWaitTimer != 0 && AutoIncWaitTimer + AUTOINC_WAIT < tNow && AutoIncTimer == 0) ||
       (AutoIncTimer != 0 && AutoIncTimer + AUTOINC_SPEED < tNow) )
  {
      AutoIncWaitTimer = tNow + AUTOINC_WAIT;
      AutoIncTimer = tNow + AUTOINC_SPEED;
      if ( AutoIncUp ==1 ) {
        if (Profiles[ProfileActive].tbuttons[M_EDIT_Btn] == 0) { // keyboard
          KButtonMappingMod(1);
        } else {
          JButtonMappingMod(1);
        }
      } else {
        if (Profiles[ProfileActive].tbuttons[M_EDIT_Btn] == 0) { // keyboard
          KButtonMappingMod(0);
        } else {
          JButtonMappingMod(0);
        }
      }
  }

  // Edit mode On/Off handling
  // start edit mode
  if ( EditToggleTimer > 0 && EditToggleTimer + EDIT_ON_TIME < tNow && Mode == M_START) {
    Mode = M_EDIT;
    M_EDIT_Step = 1; // entering editing mode
    EditToggleTimer = 0;
  }
  // leaving edit mode
  else if ( EditToggleTimer > 0 && EditToggleTimer + EDIT_OFF_TIME < tNow && Mode == M_EDIT) {
    SaveConfig();
    Mode = M_START;
    EditToggleTimer = 0;
  }

  // Display On/Off handling
  // turn screen off
  if ( DisplayToggleTimer > 0 && DisplayToggleTimer + DISPLAY_OFF_TIME < tNow && Mode == M_START) {
    DisplayToggleTimer = 0;
    Mode = M_RUNNING;
    display.clearDisplay();
    display.display();
  }
  // turn screen back on when the button was hold down long enough
  else if ( DisplayToggleTimer > 0 && DisplayToggleTimer + DISPLAY_ON_TIME < tNow && Mode == M_RUNNING) {
    DisplayToggleTimer = 0;
    Mode = M_START;
    ActiveScreen = 0;
    tNextScreen = 0;
  }

  // Option On/Off handling
  if ( OptionToggleTimer > 0 && OptionToggleTimer + OPTION_ON_TIME < tNow && Mode == M_START) {
    OptionToggleTimer = 0;
    Mode = M_OPTION;
    M_OPTION_Modify = 1;
    M_OPTION_Step = 1; // entering editing mode
  }
  else if ( OptionToggleTimer > 0 && OptionToggleTimer + OPTION_OFF_TIME < tNow && Mode == M_OPTION) {
    OptionToggleTimer = 0;
    Mode = M_START;
    ActiveScreen = 0;
    tNextScreen = 0;
  }
}

void JoyPadHandling(byte btn, bool set1) {
  int hatmod = 0;
  int btnset = 0;
  if ( set1 == 1 ) { hatmod = 1; btnset = 1; }
  joyupd = true;
/*
    Serial.print("button: ");
    Serial.print(btn);
    Serial.print(", translated: ");
    Serial.println(Profiles[ProfileActive].jbuttons[btn]);
*/
    switch (Profiles[ProfileActive].jbuttons[btn]) {
      case 18:  DUp = 1 * hatmod;
                break;
      case 19:  DDown = 1 * hatmod;
                break;
      case 20:  DLeft = 1 * hatmod;
                break;
      case 21:  DRight = 1 * hatmod;
                break;
      default: 
              Joystick.button(Profiles[ProfileActive].jbuttons[btn], btnset);
               break;
    }

    if ( DUp == 1 ) JoyHat += 2;
    if ( DDown == 1 ) JoyHat += 3;
    if ( DLeft == 1 ) JoyHat += 6;
    if ( DRight == 1 ) JoyHat += 12;

    switch(JoyHat) {
      case 2: Joystick.hat(0); break; // up
      case 3: Joystick.hat(180); break; // down
      case 6: Joystick.hat(270); break; // left
      case 8: Joystick.hat(315); break; // up-left
      case 9: Joystick.hat(225); break; // down-left
      case 12: Joystick.hat(90); break; // right
      case 14: Joystick.hat(45); break; // up-right
      case 15: Joystick.hat(135); break; // down-right
      default: Joystick.hat(-1);
    }
/*
          0
          UP
      315    45
 270 LT   -1   RT 90
      225    135
          DN
          180
*/
}

// handling modification of joypad button mapping
void JButtonMappingMod(bool up) {
  if ( up == 1 ) {
    if ( Profiles[ProfileActive].jbuttons[M_EDIT_Btn] < 21 ) Profiles[ProfileActive].jbuttons[M_EDIT_Btn]++;
    else Profiles[ProfileActive].jbuttons[M_EDIT_Btn] = 1;
  } else {
    if ( Profiles[ProfileActive].jbuttons[M_EDIT_Btn] > 1 ) Profiles[ProfileActive].jbuttons[M_EDIT_Btn]--;
    else Profiles[ProfileActive].jbuttons[M_EDIT_Btn] = 21;
  }
}

// handling modification of keyboard key mapping
void KButtonMappingMod(bool up) {
  if ( up == 1 ) {
    if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] > 47 && Profiles[ProfileActive].kbuttons[M_EDIT_Btn] < 57) Profiles[ProfileActive].kbuttons[M_EDIT_Btn]++; // 0..9
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] == 57 ) Profiles[ProfileActive].kbuttons[M_EDIT_Btn] = 97; // a
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] > 96 && Profiles[ProfileActive].kbuttons[M_EDIT_Btn] < 122) Profiles[ProfileActive].kbuttons[M_EDIT_Btn]++; // a..z
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] == 122 ) Profiles[ProfileActive].kbuttons[M_EDIT_Btn] = 48; // 0
  } else {
    if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] > 48 && Profiles[ProfileActive].kbuttons[M_EDIT_Btn] < 58) Profiles[ProfileActive].kbuttons[M_EDIT_Btn]--; // 0..9
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] == 48 ) Profiles[ProfileActive].kbuttons[M_EDIT_Btn] = 122; // z
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] > 97 && Profiles[ProfileActive].kbuttons[M_EDIT_Btn] < 123) Profiles[ProfileActive].kbuttons[M_EDIT_Btn]--; // a..z
    else if ( Profiles[ProfileActive].kbuttons[M_EDIT_Btn] == 97 ) Profiles[ProfileActive].kbuttons[M_EDIT_Btn] = 57; // 0
  }
}

void ShowProfileNr(bool simple) {
  display.setCursor(0,0);
  display.print("Profile #");
  display.print(ProfileActive+1);
  display.setCursor(65,0);
  if ( simple == false ) {
    if ( ShowData == BUTTONS_UPPER ) {
      display.print("Pad Map");
    } 
    else if ( ShowData == BUTTONS_LOWER ) {
      display.print("Screen Map");
    }
  }
}

void ShowProfileData(byte id) {
  if ( ShowData == BUTTONS_UPPER ) ShowProfileBtnCtrl(id); // show upper button mapping
  else if ( ShowData == BUTTONS_LOWER ) ShowProfileBtnScr(id); // show lower button mapping
}

//show lower button mapping, buttons located around display
void ShowProfileBtnScr(byte id) {
  byte bid; // button id (position in array)
  display.setTextSize(1);
  display.setTextColor(WHITE);

  // screen rectangle
  display.drawRect(35, 16, 58, 34, WHITE);

  // oled button left
  bid = 9;
  display.setCursor(8,34);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(20,34);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(Profiles[id].jbuttons[bid]); 
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(20,34);
    display.print("."); display.print("j");
  }
  // oled button right
  bid = 6;
  display.setCursor(98,34);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(110,34);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(110,34);
    display.print("."); display.print("j");
  }
  
  // oled button bottom-left
  bid = 8;
  display.setCursor(8,54);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(20,54);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(20,54);
    display.print("."); display.print("j");
  }

  // oled button bottom-center
  bid = 5;
  display.setCursor(56,54);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(68,54);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(68,54);
    display.print("."); display.print("j");
  }

  // oled button bottom-right
  bid = 7;
  display.setCursor(98,54);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(110,54);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(110,54);
    display.print("."); display.print("j");
  }
  
  display.display();
}

//show upper button mapping, buttons located around sticks / rear of SnapT
void ShowProfileBtnCtrl(byte id) {
  byte bid; // button id (position in array)
  display.setTextSize(1);
  display.setTextColor(WHITE);

  // backside button graphic 
  display.drawLine(35, 16, 90, 16, WHITE); // ----
  display.drawLine(35, 16, 39, 24, WHITE); //  BSlash
  display.drawLine(39, 24, 39, 32, WHITE); //  |
  display.drawLine(39, 32, 86, 32, WHITE); //   --
  display.drawLine(86, 24, 86, 32, WHITE); //  |
  display.drawLine(90, 16, 86, 24, WHITE); //   /

  // left stick button graphic
  display.drawLine( 4, 38, 54, 38, WHITE); // ----
  
  display.drawLine( 4, 38,  4, 43, WHITE); // |
  display.drawLine( 4, 43,  6, 48, WHITE); // .|
  display.drawLine( 6, 48, 12, 50, WHITE); // .-
  display.drawLine(12, 50, 30, 51, WHITE); // --
  
  display.drawLine(54, 38, 54, 43, WHITE); // |
  display.drawLine(54, 43, 52, 48, WHITE); // |
  display.drawLine(52, 48, 46, 50, WHITE); // .-
  display.drawLine(46, 50, 30, 51, WHITE); // --

  // right stick button graphic (+64)+5
  display.drawLine( 73, 38,123, 38, WHITE); // ----
  
  display.drawLine( 73, 38, 73, 43, WHITE); // |
  display.drawLine( 73, 43, 77, 48, WHITE); // .|
  display.drawLine( 77, 48, 81, 50, WHITE); // .-
  display.drawLine( 81, 50, 99, 51, WHITE); // --
  
  display.drawLine(123, 38, 123, 43, WHITE); // |
  display.drawLine(123, 43, 121, 48, WHITE); // |
  display.drawLine(121, 48, 115, 50, WHITE); // .-
  display.drawLine(115, 50,  99, 51, WHITE); // --
  
  

  // rear top-left button
  bid = 14;
  display.setCursor( 8,16);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(20,16);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(Profiles[id].jbuttons[bid]); 
    display.setCursor(20,16);
    display.print("."); display.print("j");
  }
  // rear top-right button
  display.setCursor(94,16);
  bid=0;
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(106,16);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(Profiles[id].jbuttons[bid]); 
    display.setCursor(106,16);
    display.print("."); display.print("j");
  }
  // rear bottom-left button
  bid = 13;
  display.setCursor( 8,24);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(20,24);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(Profiles[id].jbuttons[bid]); 
    display.setCursor(20,24);
    display.print("."); display.print("j");
  }
  // rear bottom-right button
  bid = 1;
  display.setCursor(94,24);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(106,24);
    display.print("."); display.print("k");
  } else { // joystick
    display.print(Profiles[id].jbuttons[bid]); 
    display.setCursor(106,24);
    display.print("."); display.print("j");
  }

  // left-stick-button left
  bid = 10;
  display.setCursor(8,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(19,40); display.print("|"); 
    display.setCursor(11,44); display.print("."); 
    display.setCursor(13,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(19,40); display.print("|"); 
    display.setCursor(11,44); display.print("."); 
    display.setCursor(13,54); display.print("j"); 
  }
  // left-stick-button mid
  bid = 11;
  display.setCursor(24,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(35,40); display.print("|"); 
    display.setCursor(27,44); display.print("."); 
    display.setCursor(29,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(35,40); display.print("|"); 
    display.setCursor(27,44); display.print("."); 
    display.setCursor(29,54); display.print("j"); 
  }
  // left-stick-button right
  bid = 12;
  display.setCursor(40,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(51,40); display.print(" "); 
    display.setCursor(43,44); display.print("."); 
    display.setCursor(45,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(51,40); display.print("|"); 
    display.setCursor(43,44); display.print("."); 
    display.setCursor(45,54); display.print("j"); 
  }

  // right-stick-button left
  bid = 2;
  display.setCursor(78,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(89,40); display.print("|"); 
    display.setCursor(81,44); display.print("."); 
    display.setCursor(83,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(89,40); display.print("|"); 
    display.setCursor(81,44); display.print("."); 
    display.setCursor(83,54); display.print("j"); 
  }
  // right-stick-button mid
  bid = 3;
  display.setCursor(94,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(105,40); display.print("|"); 
    display.setCursor(97,44); display.print("."); 
    display.setCursor(99,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(105,40); display.print("|"); 
    display.setCursor(97,44); display.print("."); 
    display.setCursor(99,54); display.print("j"); 
  }
  // right-stick-button right
  bid = 4;
  display.setCursor(110,40);
  if ( Profiles[id].tbuttons[bid] == 0 ) { // keyboard 
    display.print(Profiles[id].kbuttons[bid]); 
    display.setCursor(121,40); display.print(" "); 
    display.setCursor(113,44); display.print("."); 
    display.setCursor(115,54); display.print("k"); 
  } else { // joystick
    display.print(te2t2buttons[Profiles[id].jbuttons[bid]]);
    display.setCursor(121,40); display.print(" "); 
    display.setCursor(113,44); display.print("."); 
    display.setCursor(115,54); display.print("j"); 
  }

  display.display();
}

void LoadConfig() {
  byte value;
  value=EEPROM.read(0); // 0 : Lastused profile
  ProfileActive = value;
  if ( ProfileActive < 0 || ProfileActive > 3 ) {
    ProfileActive = 0; // limit profile number to valid values
  }
  ProfilePresets();

  value=EEPROM.read(1); // Button naming scheme 0 = ps4, 1 = xbox1,..
  EE_ButtonNames = value;
  if ( EE_ButtonNames < 0 || EE_ButtonNames > 1 ) {
    EE_ButtonNames = 0; // limit to valid values
  }

  value=EEPROM.read(2); // Show button numbers too ?
  EE_ButtonNumber = value;
  if ( EE_ButtonNumber < 0 || EE_ButtonNumber > SNAPT_BTNNR_MAX-1 ) {
    EE_ButtonNumber = 0; // limit to valid values
  }

  value=EEPROM.read(3); // when to send button/keypresses to the Titan Two (0=Running,1=Running&Edit)
  EE_OutputOnModes = value;
  if ( EE_OutputOnModes < 0 || EE_OutputOnModes > 3 ) {
    EE_OutputOnModes = 0; // limit to valid values
  }

  // load profiles
  Profile ptmp;
  for ( int loadpi = 0; loadpi < 4; loadpi++ ) {
    profiledata_ee_read(&ptmp,loadpi);
    // check profile, if valid store it in array
    if ( profiledata_validate(&ptmp) ) profiledata_activate(&ptmp,loadpi);
  }

}


void profiledata_ee_update(byte pidx) {
  byte pi=0;
  for ( int ei=10+(pidx*70); ei< 25+(pidx*70); ei++ ) {
    EEPROM.update(ei,Profiles[pidx].tbuttons[pi]);
    EEPROM.update(ei+20,Profiles[pidx].kbuttons[pi]);
    EEPROM.update(ei+40,Profiles[pidx].jbuttons[pi]);
    pi++;
  }
}

void profiledata_ee_read(Profile *vprof, byte loadpi) {
  byte pi=0;
  for ( int ei=10+(loadpi*70); ei< 25+(loadpi*70); ei++ ) {
    vprof->tbuttons[pi]=EEPROM.read(ei);
    vprof->kbuttons[pi]=EEPROM.read(ei+20);
    vprof->jbuttons[pi]=EEPROM.read(ei+40);
    pi++;
  }
}

bool profiledata_validate(Profile *vprof) {
  bool valid = true;
  for (int i=0;i<15;i++) {
    if ( vprof->tbuttons[i] < 0 || vprof->tbuttons[i] > 1 ) valid = false;
    if ( vprof->kbuttons[i] < 47 || vprof->kbuttons[i] > 122 ) valid = false;
    if ( vprof->jbuttons[i] < 0 || vprof->jbuttons[i] > 21 ) valid = false;
  }
  return valid;
}

void profiledata_activate(Profile *vprof, byte pid) {
  for (int i=0;i<15;i++) {
     Profiles[pid].tbuttons[i] = vprof->tbuttons[i];
     Profiles[pid].kbuttons[i] = vprof->kbuttons[i];
     Profiles[pid].jbuttons[i] = vprof->jbuttons[i];
  }
}

void SaveConfig() {
  EEPROM.update(0,ProfileActive); // last used profile
  EEPROM.update(1,EE_ButtonNames); // show button names as PS4 or XBOX
  EEPROM.update(2,EE_ButtonNumber); // show button numbers as prefix for button names ?
  EEPROM.update(3,EE_OutputOnModes); // when to send button/keypresses to the Titan Two

  // profiles
  for ( int savepi = 0; savepi < 4; savepi++ ) {
    profiledata_ee_update(savepi);
  }
}

void ProfilePresets() {
  //byte i;
  
  // profile are already set in struct, only change the one that should be different than default
  // test profile 1 : buttons 1..15
  
  /*
  for ( i=0;i<15;i++) {
    Profiles[1].jbuttons[i]=i+1;
    Profiles[1].tbuttons[i]=1;
  }
  */
  /*
  for ( i=0;i<15;i++) {
    static int btn = 15;
    Profiles[2].jbuttons[i]=btn;
    Profiles[2].tbuttons[i]=1;
    btn++;
  }
  */
  /*
  for ( i=0;i<15;i++) {
    static int btn = 28;
    Profiles[3].jbuttons[i]=btn;
    Profiles[3].tbuttons[i]=1;
    btn++;
  }
  */
}

